import inspect
import logging
from typing import List, Callable, Iterable, Dict

import pandas as pd
import numpy as np

log = logging.getLogger(__name__)


class FeaturesAndLabels(object):

    def __init__(self,
                 features: List[str],
                 labels: List[str],
                 target_columns: List[str] = None,
                 loss_column: str = None,
                 feature_lags: Iterable[int] = None,
                 lag_smoothing: Dict[int, Callable[[pd.Series], pd.Series]] = None,
                 **kwargs):
        self.features = features
        self.labels = labels
        self.target_columns = target_columns
        self.loss_column = loss_column
        self.feature_lags = feature_lags
        self.lag_smoothing = lag_smoothing
        self.len_feature_lags = sum(1 for _ in feature_lags) if feature_lags is not None else 1
        self.expanded_feature_length = len(features) * self.len_feature_lags if feature_lags is not None else len(features)
        self.kwargs = kwargs
        log.info(f'number of features, lags and total: {self.len_features()}')

    def len_features(self):
        return len(self.features), self.len_feature_lags, self.expanded_feature_length

    def len_labels(self):
        return len(self.labels)

    def get_feature_names(self):
        if self.feature_lags is not None:
            return np.array([[f'{feat}_{lag}'
                              for feat in self.features]
                             for lag in self.feature_lags], ndmin=2)
        else:
            return self.features

    def __getitem__(self, item):
        if isinstance(item, tuple) and len(item) == 2:
            return self.kwargs[item[0]] if item[0] in self.kwargs else item[1]
        else:
            return self.kwargs[item] if item in self.kwargs else None

    def __repr__(self):
        return f'FeaturesAndLabels({self.features},{self.labels},{self.target_columns},{self.loss_column},' \
               f'{self.feature_lags},{self.lag_smoothing},{self.probability_cutoff}) #{len(self.features)} ' \
               f'features expand to {self.expanded_feature_length}'

    def __hash__(self):
        return hash(self.__id__())

    def __eq__(self, other):
        return self.__id__() == other.__id__()

    def __id__(self):
        import dill  # only import if really needed
        smoothers = {feature: inspect.getsource(smoother) for feature, smoother in self.lag_smoothing.items()}
        return f'{self.features},{self.labels},{dill.dumps(self.feature_lags)},{smoothers}'

    def __str__(self):
        return self.__repr__()




from .models import Model
from .summary import Summary


class Fit(object):

    def __init__(self,
                 model: Model,
                 training_summary: Summary,
                 test_summary: Summary):
        self.model = model
        self.training_summary = training_summary
        self.test_summary = test_summary

    def values(self):
        return self.model, self.training_summary, self.test_summary

    def _repr_html_(self):
        return self._html_()._repr_html_()

    def _html_(self):
        # only import it needed
        from vdom.helpers import div, table, tr, td, tbody, thead, th

        model = self.model.__repr__()
        if model is None:
            model = str(self.model)

        return div(
            table(
                thead(
                    tr(
                        th("Training Data", style={'text-align': 'left'}),
                        th("Test Data", style={'text-align': 'right'})
                    )
                ),
                tbody(
                    tr(
                        td(self.training_summary._html_()),
                        td(self.test_summary._html_())
                    ),
                    tr(
                        td(
                            model,
                            colspan="2"
                        )
                    )
                ),
                style={'width': '100%'}
            ),
            style={'width': '100%', 'float': 'left'}
        )


import logging
import sys
from typing import Tuple, Union

import numpy as np
import pandas as pd

from ..model.summary import Summary

log = logging.getLogger(__name__)


class RegressionSummary(Summary):

    def __init__(self,
                 y_true: np.ndarray,
                 y_prediction: np.ndarray,
                 index: np.ndarray,
                 loss: pd.Series = None):
        self.y_true = y_true
        self.y_prediction = y_prediction.ravel() if len(y_prediction.shape) > 1 else y_prediction
        self.index = index
        self.loss = loss

    # TODO add some statistics



import logging
from typing import Callable, Tuple

import numpy as np
import pandas as pd

from .summary import RegressionSummary
from ..model.fit import Fit
from ..model.fitter import _fit, _backtest, _predict
from ..model.models import Model
from ..error.functions import mse as _mse
log = logging.getLogger(__name__)


def fit_regressor(df: pd.DataFrame,
                  model_provider: Callable[[int], Model],
                  test_size: float = 0.4,
                  cross_validation: Tuple[int, Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]] = None,
                  cache_feature_matrix: bool = False,
                  test_validate_split_seed = 42,
                  summary_printer: Callable[[np.ndarray, np.ndarray, np.ndarray], None] = None
                  ) -> Fit:

    model, train, test, index = _fit(df,
                                     model_provider,
                                     test_size = test_size,
                                     cross_validation = cross_validation,
                                     cache_feature_matrix = cache_feature_matrix,
                                     test_validate_split_seed = test_validate_split_seed,
                                     summary_printer = summary_printer)

    # assemble the result objects
    features_and_labels = model.features_and_labels
    loss = df[features_and_labels.loss_column] if features_and_labels.loss_column is not None else None

    training_summary = RegressionSummary(train[1], model.predict(train[0]), index[0], loss)
    test_summary = RegressionSummary(test[1], model.predict(test[0]), index[1], loss)
    return Fit(model, training_summary, test_summary)


def backtest_regressor(df: pd.DataFrame, model: Model) -> None:
    x, y, y_hat, index = _backtest(df, model)

    features_and_labels = model.features_and_labels
    loss = df[features_and_labels.loss_column if features_and_labels.loss_column is not None else []]

    return RegressionSummary(y, y_hat, index, loss)


def regress(df: pd.DataFrame, model: Model, tail: int = None) -> pd.DataFrame:
    dff = _predict(df, model, tail)

    # get labels calculate error
    error_function = model[("error", _mse)]
    dff["error"] = error_function(df[model.features_and_labels.labels], dff[[col for col in dff if col.startswith('prediction')]])

    return dff



from time import perf_counter as pc
from typing import Callable, Dict, Iterable, Any, List

import numpy as np
from sklearn.model_selection import KFold
from sklearn.utils.validation import _num_samples, check_random_state, indexable


def log_with_time(log_statement: Callable[[], None]):
    log_statement()
    return pc()


def unfold_parameter_space(parameter_space: Dict[str, Iterable], parameters: Dict[str, Any]) -> List[Dict]:
    if len(parameter_space) > 0:
        # more parameters need to be unfolded
        parameter, space = parameter_space.popitem()
        return list(np.array([unfold_parameter_space(parameter_space.copy(), {**parameters, parameter: argument}) for
                              argument in space]).flat)
    else:
        return parameters


class KFoldBoostRareEvents(KFold):

    def __init__(self, n_splits='warn', shuffle=False, random_state=None):
        super().__init__(n_splits, shuffle, random_state)

    def split(self, X, y=None, groups=None):
        n_samples = _num_samples(X)
        indices = np.arange(n_samples)
        rare_event_indices = indices[y >= 0.999]

        for f, (train_idx, test_idx) in enumerate(super().split(X, y, groups)):
            yield np.hstack([train_idx, rare_event_indices]), np.hstack([test_idx, rare_event_indices])



import logging
import sys
from typing import Tuple, Union

import numpy as np
import pandas as pd

from ..model.summary import Summary

log = logging.getLogger(__name__)


class ClassificationSummary(Summary):

    def __init__(self,
                 y_true: np.ndarray,
                 y_prediction: np.ndarray,
                 index: np.ndarray,
                 loss: pd.Series = None,
                 probability_cutoff: float = 0.5):
        self.y_true = y_true
        self.y_prediction = y_prediction.ravel() if len(y_prediction.shape) > 1 else y_prediction
        self.index = index
        self.loss = loss
        self.probability_cutoff = probability_cutoff
        self.confusion_matrix = self._confusion_matrix_indices()

        # immediately log some fit quality measures
        ratios = self.get_ratios()
        log.info(f"FN Ratio = {ratios[0]}, FP Ratio = {ratios[1]}")

    def set_probability_cutoff(self, probability_cutoff: float = 0.5):
        self.probability_cutoff = probability_cutoff
        self.confusion_matrix = self._confusion_matrix_indices()

    def _confusion_matrix_indices(self):
        index = self.index
        truth = self.y_true
        pred = self.y_prediction
        co = self.probability_cutoff

        try:
            confusion = np.array([[index[(truth == True) & (pred > co)], index[(truth == False) & (pred > co)]],
                                  [index[(truth == True) & (pred <= co)], index[(truth == False) & (pred <= co)]]])

            if len(confusion[0, 0]) <= 0:
                log.warning("Very bad fit with 0 TP, which leads to problems in the plot")

            return confusion
        except:
            print(f"shapes: y_true: {self.y_true.shape}, y_pred: {self.y_prediction.shape}, index: {self.index.shape}")
            print("Unexpected error:", sys.exc_info()[0])
            return None

    def get_ratios(self):
        cm = self.confusion_count()
        return cm[0,0] / (cm[1,0]  + 1), cm[0,0] / (cm[0,1]  + 1)

    def plot_backtest(self,
                      y: pd.Series = None,
                      size: Union[int, pd.Series] = None,
                      figsize: Tuple[int, int] = (16, 6)):
        # only import if required
        import seaborn as sns
        import matplotlib.pyplot as plt
        from pandas.plotting import register_matplotlib_converters

        # get rid of deprecation warning
        register_matplotlib_converters()

        # check value for back test
        if self.loss is None and y is None:
            raise ValueError("No loss column defined, whether in FeaturesAndLabels nor in plot_backtest")

        # scatter plot where confusion squares are the colors, the loss is the size
        y = y if y is not None \
                else self.loss.loc[self.index] if isinstance(self.loss, pd.Series) \
                    else self.loss[self.loss.columns[0]].loc[self.index]

        color = pd.Series(0, index=y.index)
        color.loc[self.confusion_matrix[0, 0]] = 1
        color.loc[self.confusion_matrix[1, 0]] = 2

        # get colors from: https://xkcd.com/color/rgb/
        fig, ax = plt.subplots(figsize=figsize)
        ax.set_ylim([y.min() * 1.1, 1])

        scatt =  sns.scatterplot(x=y.index,
                                 y=y,
                                 ax=ax,
                                 size=size if size is not None else y * -1,
                                 hue=color,
                                 palette=[sns.xkcd_rgb['white'], sns.xkcd_rgb['pale green'], sns.xkcd_rgb['cerise']])

        bar = sns.lineplot(x=y.index, y=self.y_prediction, ax=ax)
        plt.hlines(self.probability_cutoff, y.index.min(), y.index.max(), color=sns.xkcd_rgb['silver'])

        plt.close()
        return fig

    def confusion_loss(self):
        cm = self.confusion_matrix
        df = self.loss
        return np.array([[df.loc[cm[0, 0]].sum(), df.loc[cm[0, 1]].sum()],
                         [df.loc[cm[1, 0]].sum(), df.loc[cm[1, 1]].sum()]])

    def confusion_count(self):
        return np.array([
            [len(self.confusion_matrix[0, 0]), len(self.confusion_matrix[0, 1])],
            [len(self.confusion_matrix[1, 0]), len(self.confusion_matrix[1, 1])],
        ])

    def _repr_html_(self):
        return self._html_()._repr_html_()

    def _html_(self, width: str = '100%'):
        # only import it needed
        from vdom.helpers import div, p, img, table, tr, td, tbody, thead, th
        import matplotlib.pyplot as plt
        import base64
        import io

        if self.confusion_count()[0, 0] <= 0:
            return p('very bad fit with 0 TP!')

        image = None
        if self.loss is not None:
            with io.BytesIO() as f:
                fig = self.plot_backtest()
                fig.savefig(f, format="png", bbox_inches='tight')
                image = base64.encodebytes(f.getvalue()).decode("utf-8")
                plt.close(fig)

        cmc = self.confusion_count()
        cml = self.confusion_loss() if self.loss is not None else np.array([[0, 0], [0, 0]])

        return div(
            table(
                thead(
                    tr(
                        th("Classification Count", style={'text-align': 'left'}),
                        th("Classification Loss", style={'text-align': 'right'})
                    )
                ),
                tbody(
                    tr(
                        td(self._matrix_table(cmc)),
                        td(self._matrix_table(cml), style={'float': 'right'})
                    ),
                    tr(
                        td(
                            img(src=f'data:image/png;base64,{image}', style={'width': '100%'}) if image is not None else "",
                            colspan='2'
                        )
                    )
                ),
                style={'width': '100%'}
            ), style={'width': width}
        )

    def _matrix_table(self, mx: np.array):
        from vdom.helpers import table, tr, td, tbody, thead
        row_label = [[td("True")], [td("False")]]
        colors = [['green', 'orange'],
                  ['red', 'grey']]

        return table(
            thead(
                tr(
                    td("Prediction / Truth"),
                    td("True"),
                    td("False")
                )
            ),
            tbody(
                [tr(
                    row_label[row] + [td(
                        f'{mx[row, col]: .2f}', style={'color': colors[row][col]})
                        for col in range(mx.shape[1])])
                    for row in range(mx.shape[0])]
            )
        )

    def __len__(self):
        return len(self.y_true)

    def __str__(self) -> str:
        return f'\n{len(self.confusion_matrix[0,0])}\t{len(self.confusion_matrix[0,1])}' \
               f'\n{len(self.confusion_matrix[1,0])}\t{len(self.confusion_matrix[1,1])}'





import logging
from typing import Callable, Tuple

import numpy as np
import pandas as pd

from ..classification.summary import ClassificationSummary
from ..model.fit import Fit
from ..model.fitter import _fit, _backtest, _predict
from ..model.models import Model

log = logging.getLogger(__name__)


def fit_classifier(df: pd.DataFrame,
                   model_provider: Callable[[int], Model],
                   test_size: float = 0.4,
                   cross_validation: Tuple[int, Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]] = None,
                   cache_feature_matrix: bool = False,
                   test_validate_split_seed = 42,
                   summary_printer: Callable[[np.ndarray, np.ndarray, np.ndarray], None] = None
                   ) -> Fit:

    model, train, test, index = _fit(df,
                                     model_provider,
                                     test_size = test_size,
                                     cross_validation = cross_validation,
                                     cache_feature_matrix = cache_feature_matrix,
                                     test_validate_split_seed = test_validate_split_seed,
                                     summary_printer = summary_printer)

    # assemble the result objects
    features_and_labels = model.features_and_labels
    cutoff = model[("probability_cutoff", 0.5)]

    loss = df[features_and_labels.loss_column] if features_and_labels.loss_column is not None else None
    training_classification = ClassificationSummary(train[1], model.predict(train[0]), index[0], loss, cutoff)
    test_classification = ClassificationSummary(test[1], model.predict(test[0]), index[1], loss, cutoff)
    return Fit(model, training_classification, test_classification)


def backtest_classifier(df: pd.DataFrame, model: Model) -> ClassificationSummary:
    x, y, y_hat, index = _backtest(df, model)

    features_and_labels = model.features_and_labels
    loss = df[features_and_labels.loss_column if features_and_labels.loss_column is not None else []]
    return ClassificationSummary(y, y_hat, index, loss, model[("probability_cutoff", 0.5)])


def classify(df: pd.DataFrame, model: Model, tail: int = None) -> pd.DataFrame:
    dff = _predict(df, model, tail)

    # return result
    dff["prediction_proba"] = dff["prediction"]
    dff["prediction"] = dff["prediction_proba"] > model[("probability_cutoff", 0.5)]
    return dff



import numpy as np

from unittest import TestCase
from sklearn.metrics import confusion_matrix
from pandas_ml_utils.classification.summary import ClassificationSummary


class TestClassificationSummary(TestCase):

    def test_confusion_matrix_indices(self):
        index = np.array([1, 2, 3, 4, 5])
        pred = np.array([1.0, 0.0, 1.0, 1.0, 0.0])
        truth = np.array([True, True, True, False, False])

        cs = ClassificationSummary(truth, pred, index, None)
        cm = cs._confusion_matrix_indices()

        expected = np.array([[np.array([1, 3]), np.array([4])],
                             [np.array([2]), np.array([5])]])

        np.testing.assert_array_equal(cm[0,0], expected[0,0])
        np.testing.assert_array_equal(cm[0,1], expected[0,1])
        np.testing.assert_array_equal(cm[1,0], expected[1,0])
        np.testing.assert_array_equal(cm[1,1], expected[1,1])

        cm_skit = confusion_matrix(truth, pred)
        self.assertEqual(len(cm[0,0]), cm_skit[1,1])
        self.assertEqual(len(cm[0,1]), cm_skit[0,1])
        self.assertEqual(len(cm[1,0]), cm_skit[1,0])
        self.assertEqual(len(cm[1,1]), cm_skit[0,0])

        print(cs)


import logging
from time import perf_counter
from typing import Callable, Tuple

import numpy as np
import pandas as pd

from ..train_test_data import make_training_data, make_forecast_data
from ..utils import log_with_time
from ..model.models import Model
from ..classification.summary import ClassificationSummary

log = logging.getLogger(__name__)


def _fit(df: pd.DataFrame,
        model_provider: Callable[[int], Model],
        test_size: float = 0.4,
        cross_validation: Tuple[int, Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]] = None,
        cache_feature_matrix: bool = False,
        test_validate_split_seed = 42,
        summary_printer: Callable[[np.ndarray, np.ndarray, np.ndarray], None] = None # TODO lets provide a summary provider for the result like lambda model, df, fal, x, y, y_hat: ClassificationSummary(...)
        ) -> Tuple[Model, Tuple, Tuple, Tuple]:
    # get a new model
    model = model_provider()
    features_and_labels = model.features_and_labels

    # make training and test data sets
    x_train, x_test, y_train, y_test, index_train, index_test, min_required_data = \
        make_training_data(df,
                           features_and_labels,
                           test_size,
                           int,
                           test_validate_split_seed,
                           cache=cache_feature_matrix,
                           summary_printer=summary_printer)

    log.info(f"create model (min required data = {min_required_data}")
    model.min_required_data = min_required_data

    # fit the model
    start_performance_count = log_with_time(lambda: log.info("fit model"))
    if cross_validation is not None and isinstance(cross_validation, Tuple) and callable(cross_validation[1]):
        for fold_epoch in range(cross_validation[0]):
            # cross validation, make sure we re-shuffle every fold_epoch
            for f, (train_idx, test_idx) in enumerate(cross_validation[1](x_train, y_train)):
                log.info(f'fit fold {f}')
                model.fit(x_train[train_idx], y_train[train_idx], x_train[test_idx], y_train[test_idx])
    else:
        # fit without cross validation
        model.fit(x_train, y_train, x_test, y_test)

    log.info(f"fitting model done in {perf_counter() - start_performance_count: .2f} sec!")
    return model, (x_train, y_train), (x_test, y_test), (index_train, index_test)


def _backtest(df: pd.DataFrame, model: Model) -> ClassificationSummary:
    features_and_labels = model.features_and_labels

    # make training and test data with no 0 test data fraction
    x, _, y, _, index, _, _ = make_training_data(df, features_and_labels, 0, int)

    # predict probabilities
    y_hat = model.predict(x)
    return x, y, y_hat, index


def _predict(df: pd.DataFrame, model: Model, tail: int = None) -> pd.DataFrame:
    features_and_labels = model.features_and_labels

    if tail is not None:
        if tail <= 0:
            raise ValueError("tail must be > 0 or None")
        elif model.min_required_data is not None:
            # just use the tail for feature engineering
            df = df[-(tail + (model.min_required_data - 1)):]
        else:
            log.warning("could not determine the minimum required data from the model")

    # then re assign data frame with features only
    dff, x = make_forecast_data(df, features_and_labels)

    # first save target columns and loss column
    if features_and_labels.target_columns is not None:
        dff = dff.join(df[features_and_labels.target_columns].add_prefix("traget_"))

    if features_and_labels.loss_column is not None:
        dff["loss"] = df[features_and_labels.loss_column]

    # predict on features
    prediction = model.predict(x)
    if len(prediction.shape) > 1 and prediction.shape[1] > 1:
        for i in range(prediction.shape[1]):
            dff[f"prediction_{model.features_and_labels.labels[i]}"] = prediction[:,i]
    else:
        dff["prediction"] = prediction

    return dff



"""Augment pandas DataFrame with methods for machine learning"""
__version__ = '0.0.7'

from .pandas_utils_extension import *
from .wrappers.lazy_dataframe import *
from .datafetching.fetch_yahoo import *
from .model.models import *
from .model.features_and_Labels import *
from .classification.summary import *
from .classification.classifier import *
from .regression.regressor import *
from .train_test_data import *
from pandas.core.base import PandasObject


# add functions to pandas
# general utility functions
PandasObject.inner_join = inner_join
PandasObject.drop_re = drop_re
PandasObject.add_apply = add_apply
PandasObject.shift_inplace = shift_inplace
PandasObject.extend_forecast = extend_forecast
PandasObject.make_training_data = make_training_data

# classification functions
PandasObject.fit_classifier = fit_classifier
PandasObject.classify = classify
PandasObject.backtest_classifier = backtest_classifier

# regression functions
PandasObject.fit_regressor = fit_regressor
PandasObject.backtest_regressor = backtest_regressor
PandasObject.regress = regress

# data fetcher
setattr(pd, 'fetch_yahoo', fetch_yahoo)

# very rarely beg for love
if np.random.uniform() >= 0.99:
    print("If you like using pandas-ml-utils please show some love and star it on github: "
          "https://github.com/KIC/pandas_ml_utils")


import pandas as pd
import numpy as np
import unittest

from sklearn.model_selection import KFold

import pandas_ml_utils as pdu

from sklearn.neural_network import MLPClassifier, MLPRegressor

import logging


logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


class ComponentTest(unittest.TestCase):

    def test_fetch_yahoo(self):
        df = pd.fetch_yahoo(spy="SPY").tail()
        self.assertTrue(df["spy_Close"].sum() > 0)

    def test_fit_classifier_full(self):
        df = pd.read_csv(f'{__name__}.csv', index_col='Date')
        df['label'] = df["spy_Close"] > df["spy_Open"]

        # fit
        fit = df.fit_classifier(pdu.SkitModel(MLPClassifier(activation='tanh', hidden_layer_sizes=(60, 50), alpha=0.001,
                                                            random_state=42),
                                              pdu.FeaturesAndLabels(features=['vix_Close'], labels=['label'], target_columns=["vix_Open"], loss_column="spy_Volume")),
                                test_size=0.4,
                                test_validate_split_seed=42)

        self.assertEqual(fit.model.min_required_data, 1)
        np.testing.assert_array_equal(fit.test_summary.confusion_count(), np.array([[744, 586],
                                                                                    [655, 698]]))

        # backtest
        fitted_model = fit.model
        backtest_classification = df.backtest_classifier(fitted_model)
        np.testing.assert_array_equal(backtest_classification.confusion_count(), np.array([[1811, 1458],
                                                                                           [1657, 1780]]))

        # classify
        fitted_model = fit.model
        classified_df = df.classify(fitted_model)
        print(classified_df.tail())

        self.assertEqual(len(classified_df[classified_df["prediction"] == False]), 3437)
        self.assertTrue(classified_df["loss"].sum() > 0)
        self.assertTrue(classified_df["prediction_proba"].sum() > 0)
        self.assertTrue(classified_df["prediction_proba"].min() > 0)
        self.assertTrue(classified_df["prediction_proba"].max() < 1)
        self.assertListEqual(classified_df.columns.tolist(),
                             ["vix_Close", "traget_vix_Open", "loss", "prediction", "prediction_proba"])

        # classify tail
        fitted_model = fit.model
        classified_df = df.classify(fitted_model, 2)
        self.assertEqual(len(classified_df), 2)

    def test_fit_classifier_simple(self):
        df = pd.read_csv(f'{__name__}.csv', index_col='Date')
        df['label'] = df["spy_Close"] > df["spy_Open"]

        # fit
        fit = df.fit_classifier(pdu.SkitModel(MLPClassifier(activation='tanh', hidden_layer_sizes=(60, 50), alpha=0.001,
                                                            random_state=42),
                                              pdu.FeaturesAndLabels(features=['vix_Close'], labels=['label'])),
                                test_size=0.4,
                                test_validate_split_seed=42)

        fitted_model = fit.model
        classified_df = df.classify(fitted_model)

        self.assertListEqual(classified_df.columns.tolist(),
                             ["vix_Close", "prediction", "prediction_proba"])

    def test_fit_regressor(self):
        df = pd.read_csv(f'{__name__}.csv', index_col='Date') / 50.

        # fit
        fit = df.fit_regressor(pdu.SkitModel(MLPRegressor(activation='tanh', hidden_layer_sizes=(4, 3, 2, 1, 2, 3, 4),
                                                          random_state=42),
                                             pdu.FeaturesAndLabels(features=['vix_Open', 'vix_High', 'vix_Low', 'vix_Close'],
                                                                   labels=['vix_Open', 'vix_High', 'vix_Low', 'vix_Close'])),
                               test_size=0.4,
                               test_validate_split_seed=42)

        fitted_model = fit.model

        # backtest
        backtest_regression = df.backtest_regressor(fitted_model)
        self.assertIsNotNone(backtest_regression) # FIXME better assertion

        # regressed
        regressed = df.regress(fitted_model)
        print(regressed.tail())
        self.assertListEqual(regressed.columns.tolist(),
                             ['vix_Open', 'vix_High', 'vix_Low', 'vix_Close',
                              'prediction_vix_Open', 'prediction_vix_High', 'prediction_vix_Low', 'prediction_vix_Close',
                              'error'])

        self.assertTrue(regressed["error"].sum() > 0)
        self.assertTrue(regressed["error"].sum() > 0)
        self.assertTrue(regressed["error"].min() > 0)

    def test_cross_validation(self):
        df = pd.read_csv(f'{__name__}.csv', index_col='Date')
        df['label'] = df["spy_Close"] > df["spy_Open"]

        # KFold
        cv = KFold(n_splits=10, shuffle=False)

        # fit
        fit = df.fit_classifier(pdu.SkitModel(MLPClassifier(activation='tanh', hidden_layer_sizes=(60, 50), alpha=0.001,
                                                            random_state=42, max_iter=10),
                                              pdu.FeaturesAndLabels(features=['vix_Close'], labels=['label'],
                                                                    target_columns=["vix_Open"],
                                                                    loss_column="spy_Volume")),
                                test_size=0.4,
                                cross_validation = (2, cv.split),
                                test_validate_split_seed=42)

        self.assertEqual(fit.model.min_required_data, 1)
        np.testing.assert_array_equal(fit.test_summary.confusion_count(), np.array([[257, 169],
                                                                                    [1142, 1115]]))




#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pandas_ml_utils',
 'pandas_ml_utils.classification',
 'pandas_ml_utils.datafetching',
 'pandas_ml_utils.error',
 'pandas_ml_utils.extern',
 'pandas_ml_utils.model',
 'pandas_ml_utils.regression',
 'pandas_ml_utils.wrappers']

package_data = \
{'': ['*']}

install_requires = \
['pandas', 'cachetools', 'sortedcontainers', 'scikit-learn', 'dill']

extras_require = \
{'jupyther': ['vdom', 'matplotlib', 'seaborn'], 'stock_prices': ['yfinance']}

setup(name='pandas-ml-utils',
      version='0.0.7',
      description='Augment pandas DataFrame with methods for machine learning',
      author='KIC',
      author_email='kic@mindbusters.com',
      url='https://github.com/KIC/pandas_ml_utils',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>= 3.7',
     )



import logging
from copy import deepcopy

import dill as pickle
import numpy as np

from .features_and_Labels import FeaturesAndLabels
from ..train_test_data import reshape_rnn_as_ar

log = logging.getLogger(__name__)


class Model(object):

    @staticmethod
    def load(filename: str):
        with open(filename, 'rb') as file:
            model = pickle.load(file)
            if isinstance(model, Model):
                return model
            else:
                raise ValueError("Deserialized pickle was not a Model!")

    def __init__(self, features_and_labels: FeaturesAndLabels, **kwargs):
        self.features_and_labels = features_and_labels
        self.min_required_data: int = None
        self.kwargs = kwargs

    def __getitem__(self, item):
        if isinstance(item, tuple) and len(item) == 2:
            return self.kwargs[item[0]] if item[0] in self.kwargs else item[1]
        else:
            return self.kwargs[item] if item in self.kwargs else None

    def save(self, filename: str):
        with open(filename, 'wb') as file:
            pickle.dump(self, file)

    def fit(self, x, y, x_val, y_val) -> None:
        pass

    def predict(self, x) -> np.ndarray:
        pass

    # this lets the model itself act as a provider. However we want to use the same Model configuration
    # for different datasets (i.e. as part of MultiModel)
    def __call__(self, *args, **kwargs):
        return deepcopy(self)


class SkitModel(Model):

    def __init__(self, skit_model, features_and_labels: FeaturesAndLabels, **kwargs):
        super().__init__(features_and_labels, **kwargs)
        self.skit_model = skit_model

    def fit(self, x, y, x_val, y_val):
        self.skit_model.fit(reshape_rnn_as_ar(x), y),

    def predict(self, x):
        if callable(getattr(self.skit_model, 'predict_proba', None)):
            return self.skit_model.predict_proba(reshape_rnn_as_ar(x))[:, 1]
        else:
            return self.skit_model.predict(reshape_rnn_as_ar(x))


# TODO add Keras Model
class KerasModel(Model):
    pass

# class MultiModel(Model):
#
#     def __init__(self, model_provider: Callable[[], Model], features_and_labels: FeaturesAndLabels):
#         super().__init__(features_and_labels)
#         self.model_provider = model_provider
#
#     def fit(self, x, y, x_val, y_val) -> None:
#         pass
#
#     def predict(self, x) -> np.ndarray:
#         # we would need to return a prediction for every and each parameters dict in the parameter space
#         pass
